#ifndef MYX_BASE_ENDIAN_TYPES_HPP_
#define MYX_BASE_ENDIAN_TYPES_HPP_

#pragma once

#include <myx/backports/cpp/span.hpp>

#include <cassert>
#include <cstdint>
#include <iostream>

#ifdef QT_CORE_LIB
#include <QDebug>
#endif

// Основа здесь: https://github.com/tatewake/endian-template/

namespace myx {

namespace core {

template< typename T >
class EndianTypesBase
{
protected:
	#if defined ( _MSC_VER )
	#pragma warning ( push )
	#endif
	static T swapBytes( const T& b )
	{
		T n;
		std::span< uint8_t > sn( reinterpret_cast< uint8_t* >( &n ), sizeof( T ) );             //NOLINT
		std::span< const uint8_t > sb( reinterpret_cast< const uint8_t* >( &b ), sizeof( T ) ); //NOLINT
		std::reverse_copy( sb.begin(), sb.end(), sn.begin() );
		return( n );
	} // swapBytes


	#if defined( _MSC_VER )
	#pragma warning ( pop )
	#endif
}; // class EndianTypesBase

template< typename T >
class LittleEndianType : public EndianTypesBase< T >
{
protected:
	T m_data; //NOLINT

	static T transform( const T& b )
	{
		const uint16_t k_i = 1;
		return( ( reinterpret_cast< const char& >( k_i ) == 1 ) ? b : EndianTypesBase< T >::swapBytes( b ) ); //NOLINT
	}


public:
	// Constructors
	LittleEndianType() = default;

	// If we endian swap, it happens in two places:
	// 1. Set an OE from a PE
	// 2. Get a PE from an OE
	// Storage in
	// NOLINTNEXTLINE
	LittleEndianType( const T& b ) :
		m_data( transform( b ) )
	{}
	template < typename U > explicit LittleEndianType( U const& b ) :
		m_data( transform( T( b ) ) )
	{}
	// Storage out
	template < typename U > explicit operator U() const { return( U( transform( m_data ) ) ); }
	explicit operator T() const { return( transform( m_data ) ); }

	template < typename U > bool operator ==( U const& o ) { return( U( *this ) == o ); }
	template < typename U > bool operator !=( U const& o ) { return( U( *this ) != o ); }

	// Arithmetic assignment operators
	LittleEndianType& operator ++() /* prefix */ { *this = T( *this ) + T( 1 ); return( *this ); }
	LittleEndianType operator ++( int ) /* suffix */ { LittleEndianType t( *this );   *this = T( *this ) + T( 1 ); return( t ); }
	LittleEndianType& operator --() /* prefix */ { *this = T( *this ) - T( 1 ); return( *this ); }
	LittleEndianType operator --( int ) /* suffix */ { LittleEndianType t( *this );   *this = T( *this ) - T( 1 ); return( t ); }

	// Compound assignment operators
	LittleEndianType& operator +=( const T& b )  { *this = T( *this ) + b; return( *this ); }
	LittleEndianType& operator -=( const T& b )  { *this = T( *this ) - b; return( *this ); }
	LittleEndianType& operator *=( const T& b )  { *this = T( *this ) * b; return( *this ); }
	LittleEndianType& operator /=( const T& b )  { *this = T( *this ) / b; return( *this ); }
	LittleEndianType& operator %=( const T& b )  { *this = T( *this ) % b; return( *this ); }
	LittleEndianType& operator &=( const T& b )  { *this = T( *this ) & b; return( *this ); }
	LittleEndianType& operator |=( const T& b )  { *this = T( *this ) | b; return( *this ); }
	LittleEndianType& operator ^=( const T& b )  { *this = T( *this ) ^ b; return( *this ); }
	LittleEndianType& operator <<=( const T& b ) { *this = T( T( *this ) << b ); return( *this ); }
	LittleEndianType& operator >>=( const T& b ) { *this = T( T( *this ) >> b ); return( *this ); }
	friend std::ostream& operator <<( std::ostream& out, const LittleEndianType b ) { out << T( b ); return( out ); }
	friend std::istream& operator >>( std::istream& in, LittleEndianType& b )       { T val; in >> val; b = val; return( in ); }
	std::string toStdString() { return( std::to_string( transform( m_data ) ) ); }
	#ifdef QT_CORE_LIB
	friend QDebug& operator <<( QDebug& out, const LittleEndianType< T > b ) { out << T( b ); return( out ); }
	QString toQString() { return( QString::number( transform( m_data ) ) ); }
	#endif
}; // class LittleEndianType

template< typename T >
class BigEndianType : public EndianTypesBase< T >
{
protected:
	T m_data; //NOLINT

	static T transform( const T& b )
	{
		const uint16_t k_i = 1;
		return( ( reinterpret_cast< const char& >( k_i ) == 1 ) ? EndianTypesBase< T >::swapBytes( b ) : b ); //NOLINT
	}


public:
	// Constructors
	BigEndianType() = default;

	// If we endian swap, it happens in two places:
	// 1. Set an OE from a PE
	// 2. Get a PE from an OE
	// Storage in
	// NOLINTNEXTLINE
	BigEndianType( const T& b ) :
		m_data( transform( b ) )
	{}
	template < typename U > explicit BigEndianType( U const& b ) :
		m_data( transform( T( b ) ) )
	{}
	// Storage out
	template < typename U > explicit operator U() const { return( U( transform( m_data ) ) ); }
	explicit operator T() const { return( transform( m_data ) ); }

	template < typename U > bool operator ==( U const& o ) { return( U( *this ) == o ); }
	template < typename U > bool operator !=( U const& o ) { return( U( *this ) != o ); }

	// Arithmetic assignment operators
	BigEndianType& operator ++() /* prefix */ { *this = T( *this ) + 1; return( *this ); }
	BigEndianType operator ++( int ) /* suffix */ { BigEndianType t( *this ); *this = T( *this ) + 1; return( t ); }
	BigEndianType& operator --() /* prefix */ { *this = T( *this ) - 1; return( *this ); }
	BigEndianType operator --( int ) /* suffix */ { BigEndianType t( *this ); *this = T( *this ) - 1; return( t ); }

	// Compound assignment operators
	BigEndianType& operator +=( const T& b )  { *this = T( *this ) + b; return( *this ); }
	BigEndianType& operator -=( const T& b )  { *this = T( *this ) - b; return( *this ); }
	BigEndianType& operator *=( const T& b )  { *this = T( *this ) * b; return( *this ); }
	BigEndianType& operator /=( const T& b )  { *this = T( *this ) / b; return( *this ); }
	BigEndianType& operator %=( const T& b )  { *this = T( *this ) % b; return( *this ); }
	BigEndianType& operator &=( const T& b )  { *this = T( *this ) & b; return( *this ); }
	BigEndianType& operator |=( const T& b )  { *this = T( *this ) | b; return( *this ); }
	BigEndianType& operator ^=( const T& b )  { *this = T( *this ) ^ b; return( *this ); }
	BigEndianType& operator <<=( const T& b ) { *this = T( T( *this ) << b ); return( *this ); }
	BigEndianType& operator >>=( const T& b ) { *this = T( T( *this ) >> b ); return( *this ); }
	friend std::ostream& operator <<( std::ostream& out, const BigEndianType b ) { out << T( b ); return( out ); }
	friend std::istream& operator >>( std::istream& in, BigEndianType& b )       { T val; in >> val; b = val;    return( in ); }
	std::string toStdString() { return( std::to_string( transform( m_data ) ) ); }
	#ifdef QT_CORE_LIB
	friend QDebug& operator <<( QDebug& out, const BigEndianType< T > b ) { out << T( b ); return( out ); }
	QString toQString() { return( QString::number( transform( m_data ) ) ); }
	#endif
}; // class BigEndianType

} // namespace core

} // namespace myx

using leint16 = myx::core::LittleEndianType< int16_t >; //NOLINT
using leint32 = myx::core::LittleEndianType< int32_t >; //NOLINT
using leint64 = myx::core::LittleEndianType< int64_t >; //NOLINT

using leuint16 = myx::core::LittleEndianType< uint16_t >; //NOLINT
using leuint32 = myx::core::LittleEndianType< uint32_t >; //NOLINT
using leuint64 = myx::core::LittleEndianType< uint64_t >; //NOLINT

using lefloat  = myx::core::LittleEndianType< float >;  //NOLINT
using ledouble = myx::core::LittleEndianType< double >; //NOLINT

using beint16 = myx::core::BigEndianType< int16_t >; //NOLINT
using beint32 = myx::core::BigEndianType< int32_t >; //NOLINT
using beint64 = myx::core::BigEndianType< int64_t >; //NOLINT

using beuint16 = myx::core::BigEndianType< uint16_t >; //NOLINT
using beuint32 = myx::core::BigEndianType< uint32_t >; //NOLINT
using beuint64 = myx::core::BigEndianType< uint64_t >; //NOLINT

using befloat  = myx::core::BigEndianType< float >;  //NOLINT
using bedouble = myx::core::BigEndianType< double >; //NOLINT

#endif // MYX_BASE_ENDIAN_TYPES_HPP_
